@@ -1076,7 +1076,7 @@ bool pnpoly_mp(std::vector<mvt_geometry> const &geom, long long x, long long y)
 }
 
 std::string overzoom(std::vector<input_tile> const &tiles, int nz, int nx, int ny,
-		     int detail, int buffer,
+		     int detail_or_unspecified, int buffer,
 		     std::set<std::string> const &keep,
 		     std::set<std::string> const &exclude,
 		     std::vector<std::string> const &exclude_prefix,
@@ -1087,7 +1087,7 @@ std::string overzoom(std::vector<input_tile> const &tiles, int nz, int nx, int n
 		     std::vector<std::string> const &unidecode_data, double simplification,
 		     double tiny_polygon_size,
 		     std::vector<mvt_layer> const &bins, std::string const &bin_by_id_list,
-		     std::string const &accumulate_numeric) {
+		     std::string const &accumulate_numeric, size_t feature_limit) {
 	std::vector<source_tile> decoded;
 
 	for (auto const &t : tiles) {
@@ -1113,7 +1113,7 @@ std::string overzoom(std::vector<input_tile> const &tiles, int nz, int nx, int n
 		decoded.push_back(out);
 	}
 
-	return overzoom(decoded, nz, nx, ny, detail, buffer, keep, exclude, exclude_prefix, do_compress, next_overzoomed_tiles, demultiply, filter, preserve_input_order, attribute_accum, unidecode_data, simplification, tiny_polygon_size, bins, bin_by_id_list, accumulate_numeric);
+	return overzoom(decoded, nz, nx, ny, detail_or_unspecified, buffer, keep, exclude, exclude_prefix, do_compress, next_overzoomed_tiles, demultiply, filter, preserve_input_order, attribute_accum, unidecode_data, simplification, tiny_polygon_size, bins, bin_by_id_list, accumulate_numeric, feature_limit);
 }
 
 // like a minimal serial_feature, but with mvt_feature-style attributes
@@ -1163,14 +1163,14 @@ static void add_mean(mvt_feature &feature, mvt_layer &layer, std::string const &
 };
 
 // accumulate :sum:, :min:, :max:, and :count: versions of the specified attribute
-static void preserve_numeric(const std::string &key, const mvt_value &val,			  // numeric attribute being accumulated
-			     std::vector<std::string> &full_keys,				  // keys of feature being accumulated onto
-			     std::vector<serial_val> &full_values,				  // values of features being accumulated onto
-			     const std::string &accumulate_numeric,				  // prefix of accumulations
-			     std::set<std::string> &keys,					  // key presence in the source feature
-			     std::map<std::string, size_t> &numeric_out_field,			  // key index in the output feature
-			     std::unordered_map<std::string, accum_state> &attribute_accum_state  // accumulation state for preserve_attribute()
-) {
+static void preserve_numeric(const std::string &key, const mvt_value &val,			   // numeric attribute being accumulated
+			     std::vector<std::shared_ptr<std::string>> &full_keys,		   // keys of feature being accumulated onto
+			     std::vector<mvt_value> &full_values,				   // values of features being accumulated onto
+			     const std::string &accumulate_numeric,				   // prefix of accumulations
+			     std::set<std::string> &keys,					   // key presence in the source feature
+			     std::map<std::string, size_t> &numeric_out_field,			   // key index in the output feature
+			     std::unordered_map<std::string, accum_state> &attribute_accum_state,  // accumulation state for preserve_attribute()
+			     key_pool &key_pool) {
 	// If this is a numeric attribute, but there is also a prefix:sum (etc.) for the
 	// same attribute, we want to use that one instead of this one.
 
@@ -1213,41 +1213,37 @@ static void preserve_numeric(const std::string &key, const mvt_value &val,			  /
 				if (out_attr == numeric_out_field.end()) {
 					// not present at all, so copy our value to the prefixed output
 					numeric_out_field.emplace(prefixed, full_keys.size());
-					full_keys.push_back(prefixed);
+					full_keys.push_back(key_pool.pool(prefixed));
 
 					if (op.second == op_count) {
 						if (starting_from_accumulation) {
 							// copy our count
-							full_values.push_back(mvt_value_to_serial_val(val));
+							full_values.push_back(val);
 						} else {
 							// new count of 1
-							serial_val sv;
-							sv.type = mvt_double;
-							sv.s = "1";
-							full_values.push_back(sv);
+							full_values.push_back(mvt_value(1));
 						}
 					} else {
-						full_values.push_back(mvt_value_to_serial_val(val));
+						full_values.push_back(val);
 					}
 				} else {
 					// exists unprefixed, so copy it, and then accumulate on our value
 					numeric_out_field.emplace(prefixed, full_keys.size());
-					full_keys.push_back(prefixed);
+					full_keys.push_back(key_pool.pool(prefixed));
 
 					if (op.second == op_count) {
-						serial_val sv;
-						sv.type = mvt_double;
+						mvt_value v;
 						if (starting_from_accumulation) {
 							// sum our count onto the existing 1
-							sv.s = std::to_string(1 + mvt_value_to_long_long(val));
+							v = mvt_value(1 + mvt_value_to_long_long(val));
 						} else {
 							// sum our 1 onto the existing 1
-							sv.s = "2";
+							v = mvt_value(2);
 						}
-						full_values.push_back(sv);
+						full_values.push_back(v);
 					} else {
 						full_values.push_back(full_values[out_attr->second]);
-						preserve_attribute(op.second, prefixed, mvt_value_to_serial_val(val), full_keys, full_values, attribute_accum_state);
+						preserve_attribute(op.second, prefixed, val, full_keys, full_values, attribute_accum_state, key_pool);
 					}
 				}
 			} else {
@@ -1255,12 +1251,12 @@ static void preserve_numeric(const std::string &key, const mvt_value &val,			  /
 				if (op.second == op_count) {
 					if (starting_from_accumulation) {
 						// sum our count onto the existing count
-						full_values[prefixed_attr->second].s = std::to_string(atoll(full_values[prefixed_attr->second].s.c_str()) + mvt_value_to_long_long(val));
+						full_values[prefixed_attr->second] = mvt_value(mvt_value_to_long_long(full_values[prefixed_attr->second]) + mvt_value_to_long_long(val));
 					} else {
-						full_values[prefixed_attr->second].s = std::to_string(atoll(full_values[prefixed_attr->second].s.c_str()) + 1);
+						full_values[prefixed_attr->second] = mvt_value(mvt_value_to_long_long(full_values[prefixed_attr->second]) + 1);
 					}
 				} else {
-					preserve_attribute(op.second, prefixed, mvt_value_to_serial_val(val), full_keys, full_values, attribute_accum_state);
+					preserve_attribute(op.second, prefixed, val, full_keys, full_values, attribute_accum_state, key_pool);
 				}
 			}
 		}
@@ -1288,18 +1284,54 @@ static bool should_keep(std::string const &key,
 	return false;
 }
 
+static void handle_closepath_from_mvt(drawvec &geom) {
+	// mvt geometries close polygons with a mvt_closepath operation
+	// tippecanoe-internal geometries close polygons with a lineto to the initial point
+
+	size_t last_open = 0;
+
+	for (size_t i = 0; i < geom.size(); i++) {
+		if (geom[i].op == mvt_closepath) {
+			geom[i] = draw(mvt_lineto, geom[last_open].x, geom[last_open].y);
+		} else if (geom[i].op == mvt_moveto) {
+			last_open = i;
+		}
+	}
+}
+
 static void feature_out(std::vector<tile_feature> const &features, mvt_layer &outlayer,
 			std::set<std::string> const &keep,
 			std::set<std::string> const &exclude,
 			std::vector<std::string> const &exclude_prefix,
 			std::unordered_map<std::string, attribute_op> const &attribute_accum,
-			std::shared_ptr<std::string> const &tile_stringpool,
-			std::string const &accumulate_numeric) {
+			std::string const &accumulate_numeric,
+			key_pool &key_pool, int buffer) {
 	// Add geometry to output feature
 
+	drawvec geom = features[0].geom;
+	if (buffer >= 0) {
+		int t = features[0].t;
+
+		if (t == VT_LINE) {
+			geom = clip_lines(geom, 32 - outlayer.detail(), buffer);
+		} else if (t == VT_POLYGON) {
+			drawvec dv;
+			handle_closepath_from_mvt(geom);
+			geom = simple_clip_poly(geom, 32 - outlayer.detail(), buffer, dv, false);
+		} else if (t == VT_POINT) {
+			geom = clip_point(geom, 32 - outlayer.detail(), buffer);
+		}
+
+		geom = remove_noop(geom, t, 0);
+		if (t == VT_POLYGON) {
+			geom = clean_or_clip_poly(geom, 0, 0, false, false);
+			geom = close_poly(geom);
+		}
+	}
+
 	mvt_feature outfeature;
 	outfeature.type = features[0].t;
-	for (auto const &g : features[0].geom) {
+	for (auto const &g : geom) {
 		outfeature.geometry.emplace_back(g.op, g.x, g.y);
 	}
 
@@ -1315,13 +1347,13 @@ static void feature_out(std::vector<tile_feature> const &features, mvt_layer &ou
 
 		if (attribute_accum.size() > 0 || accumulate_numeric.size() > 0) {
 			// convert the attributes of the output feature
-			// from mvt_value to serial_val so they can have
+			// from layer references to a vector so they can have
 			// attributes from the other features of the
 			// multiplier cluster accumulated onto them
 
 			std::unordered_map<std::string, accum_state> attribute_accum_state;
-			std::vector<std::string> full_keys;
-			std::vector<serial_val> full_values;
+			std::vector<std::shared_ptr<std::string>> full_keys;
+			std::vector<mvt_value> full_values;
 			std::map<std::string, size_t> numeric_out_field;
 
 			for (size_t i = 0; i + 1 < features[0].tags.size(); i += 2) {
@@ -1329,13 +1361,13 @@ static void feature_out(std::vector<tile_feature> const &features, mvt_layer &ou
 				auto f = attribute_accum.find(key);
 				if (f != attribute_accum.end()) {
 					// this attribute has an accumulator, so convert it
-					full_keys.push_back(features[0].layer->keys[features[0].tags[i]]);
-					full_values.push_back(mvt_value_to_serial_val(features[0].layer->values[features[0].tags[i + 1]]));
+					full_keys.push_back(key_pool.pool(features[0].layer->keys[features[0].tags[i]]));
+					full_values.push_back(features[0].layer->values[features[0].tags[i + 1]]);
 				} else if (accumulate_numeric.size() > 0 && features[0].layer->values[features[0].tags[i + 1]].is_numeric()) {
 					// convert numeric for accumulation
 					numeric_out_field.emplace(key, full_keys.size());
-					full_keys.push_back(key);
-					full_values.push_back(mvt_value_to_serial_val(features[0].layer->values[features[0].tags[i + 1]]));
+					full_keys.push_back(key_pool.pool(key));
+					full_values.push_back(features[0].layer->values[features[0].tags[i + 1]]);
 				} else {
 					// otherwise just tag it directly onto the output feature
 					if (should_keep(features[0].layer->keys[features[0].tags[i]], keep, exclude, exclude_prefix)) {
@@ -1361,14 +1393,14 @@ static void feature_out(std::vector<tile_feature> const &features, mvt_layer &ou
 
 					auto f = attribute_accum.find(key);
 					if (f != attribute_accum.end()) {
-						serial_val val = mvt_value_to_serial_val(features[i].layer->values[features[i].tags[j + 1]]);
-						preserve_attribute(f->second, key, val, full_keys, full_values, attribute_accum_state);
+						mvt_value val = features[i].layer->values[features[i].tags[j + 1]];
+						preserve_attribute(f->second, key, val, full_keys, full_values, attribute_accum_state, key_pool);
 					} else if (accumulate_numeric.size() > 0) {
 						const mvt_value &val = features[i].layer->values[features[i].tags[j + 1]];
 						if (val.is_numeric()) {
 							preserve_numeric(key, val, full_keys, full_values,
 									 accumulate_numeric,
-									 keys, numeric_out_field, attribute_accum_state);
+									 keys, numeric_out_field, attribute_accum_state, key_pool);
 						}
 					}
 				}
@@ -1378,8 +1410,8 @@ static void feature_out(std::vector<tile_feature> const &features, mvt_layer &ou
 			// and tag them onto the output feature
 
 			for (size_t i = 0; i < full_keys.size(); i++) {
-				if (should_keep(full_keys[i], keep, exclude, exclude_prefix)) {
-					outlayer.tag(outfeature, full_keys[i], stringified_to_mvt_value(full_values[i].type, full_values[i].s.c_str(), tile_stringpool));
+				if (should_keep(*full_keys[i], keep, exclude, exclude_prefix)) {
+					outlayer.tag(outfeature, *full_keys[i], full_values[i]);
 				}
 			}
 
@@ -1520,13 +1552,19 @@ static std::vector<size_t> parse_ids_string(mvt_value const &v) {
 
 mvt_tile assign_to_bins(mvt_tile &features,
 			std::vector<mvt_layer> const &bins, std::string const &bin_by_id_list,
-			int z, int x, int y, int detail,
+			int z, int x, int y,
 			std::unordered_map<std::string, attribute_op> const &attribute_accum,
 			std::string const &accumulate_numeric,
 			std::set<std::string> keep,
 			std::set<std::string> exclude,
-			std::vector<std::string> exclude_prefix) {
+			std::vector<std::string> exclude_prefix,
+			int buffer) {
 	std::vector<index_event> events;
+	key_pool key_pool;
+
+	if (bins.size() == 0) {
+		return mvt_tile();
+	}
 
 	// Index bins
 	for (size_t i = 0; i < bins.size(); i++) {
@@ -1534,7 +1572,7 @@ mvt_tile assign_to_bins(mvt_tile &features,
 			long long xmin, ymin, xmax, ymax;
 			unsigned long long start, end;
 
-			get_bbox(bins[i].features[j].geometry, &xmin, &ymin, &xmax, &ymax, z, x, y, detail);
+			get_bbox(bins[i].features[j].geometry, &xmin, &ymin, &xmax, &ymax, z, x, y, bins[i].detail());
 			get_quadkey_bounds(xmin, ymin, xmax, ymax, &start, &end);
 			events.emplace_back(start, index_event::ENTER, i, j, xmin, ymin, xmax, ymax);
 			events.emplace_back(end, index_event::EXIT, i, j, xmin, ymin, xmax, ymax);
@@ -1554,7 +1592,7 @@ mvt_tile assign_to_bins(mvt_tile &features,
 					fid_to_feature.emplace(features.layers[i].features[j].id, std::make_pair(i, j));
 				}
 
-				get_bbox(features.layers[i].features[j].geometry, &xmin, &ymin, &xmax, &ymax, z, x, y, detail);
+				get_bbox(features.layers[i].features[j].geometry, &xmin, &ymin, &xmax, &ymax, z, x, y, features.layers[i].detail());
 				get_quadkey_bounds(xmin, ymin, xmax, ymax, &start, &end);
 				events.emplace_back(start, index_event::CHECK, i, j, xmin, ymin, xmax, ymax);
 			}
@@ -1565,12 +1603,11 @@ mvt_tile assign_to_bins(mvt_tile &features,
 	std::set<active_bin> active;
 
 	mvt_layer outlayer;
-	outlayer.extent = 1 << detail;
+	outlayer.extent = bins[0].extent;
 	outlayer.version = 2;
 	outlayer.name = features.layers[0].name;
 
 	std::vector<std::vector<tile_feature>> outfeatures;
-	std::shared_ptr<std::string> tile_stringpool = std::make_shared<std::string>();
 
 	for (auto &e : events) {
 		if (e.kind == index_event::ENTER) {
@@ -1629,6 +1666,10 @@ mvt_tile assign_to_bins(mvt_tile &features,
 
 			active.insert(std::move(a));
 		} else if (e.kind == index_event::CHECK) {
+			if (bin_by_id_list.size() > 0) {
+				continue;  // only bin by id, not geometrically
+			}
+
 			auto const &feature = features.layers[e.layer].features[e.feature];
 
 			if (feature.geometry.size() == 0) {
@@ -1680,7 +1721,7 @@ mvt_tile assign_to_bins(mvt_tile &features,
 		if (outfeatures[i].size() > 1) {
 			feature_out(outfeatures[i], outlayer,
 				    keep, exclude, exclude_prefix, attribute_accum,
-				    tile_stringpool, accumulate_numeric);
+				    accumulate_numeric, key_pool, buffer);
 			mvt_feature &nfeature = outlayer.features.back();
 			mvt_value val;
 			val.type = mvt_uint;
@@ -1702,7 +1743,7 @@ mvt_tile assign_to_bins(mvt_tile &features,
 }
 
 std::string overzoom(std::vector<source_tile> const &tiles, int nz, int nx, int ny,
-		     int detail, int buffer,
+		     int detail_or_unspecified, int buffer,
 		     std::set<std::string> const &keep,
 		     std::set<std::string> const &exclude,
 		     std::vector<std::string> const &exclude_prefix,
@@ -1713,15 +1754,15 @@ std::string overzoom(std::vector<source_tile> const &tiles, int nz, int nx, int
 		     std::vector<std::string> const &unidecode_data, double simplification,
 		     double tiny_polygon_size,
 		     std::vector<mvt_layer> const &bins, std::string const &bin_by_id_list,
-		     std::string const &accumulate_numeric) {
+		     std::string const &accumulate_numeric, size_t feature_limit) {
 	mvt_tile outtile;
-	std::shared_ptr<std::string> tile_stringpool = std::make_shared<std::string>();
+	key_pool key_pool;
 
 	for (auto const &tile : tiles) {
 		for (auto const &layer : tile.tile.layers) {
 			mvt_layer *outlayer = NULL;
 
-			int det = detail;
+			int det = detail_or_unspecified;
 			if (det <= 0) {
 				det = std::round(log(layer.extent) / log(2));
 			}
@@ -1840,7 +1881,10 @@ std::string overzoom(std::vector<source_tile> const &tiles, int nz, int nx, int
 
 				if (flush_multiplier_cluster) {
 					if (pending_tile_features.size() > 0) {
-						feature_out(pending_tile_features, *outlayer, keep, exclude, exclude_prefix, attribute_accum, tile_stringpool, accumulate_numeric);
+						feature_out(pending_tile_features, *outlayer, keep, exclude, exclude_prefix, attribute_accum, accumulate_numeric, key_pool, -1);
+						if (outlayer->features.size() >= feature_limit) {
+							break;
+						}
 						pending_tile_features.clear();
 					}
 				}
@@ -1854,16 +1898,16 @@ std::string overzoom(std::vector<source_tile> const &tiles, int nz, int nx, int
 				if (t == VT_POLYGON && tiny_polygon_size > 0) {
 					bool simplified_away_by_reduction = false;
 
-					geom = reduce_tiny_poly(geom, nz, detail, &still_need_simplification_after_reduction, &simplified_away_by_reduction, &accum_area, tiny_polygon_size);
+					geom = reduce_tiny_poly(geom, nz, det, &still_need_simplification_after_reduction, &simplified_away_by_reduction, &accum_area, tiny_polygon_size);
 				} else {
 					still_need_simplification_after_reduction = true;
 				}
 
 				if (simplification > 0 && still_need_simplification_after_reduction) {
 					if (t == VT_POLYGON) {
-						geom = simplify_lines_basic(geom, nz, detail, simplification, 4);
+						geom = simplify_lines_basic(geom, nz, det, simplification, 4);
 					} else if (t == VT_LINE) {
-						geom = simplify_lines_basic(geom, nz, detail, simplification, 0);
+						geom = simplify_lines_basic(geom, nz, det, simplification, 0);
 					}
 				}
 
@@ -1897,8 +1941,11 @@ std::string overzoom(std::vector<source_tile> const &tiles, int nz, int nx, int
 			}
 
 			if (pending_tile_features.size() > 0) {
-				feature_out(pending_tile_features, *outlayer, keep, exclude, exclude_prefix, attribute_accum, tile_stringpool, accumulate_numeric);
+				feature_out(pending_tile_features, *outlayer, keep, exclude, exclude_prefix, attribute_accum, accumulate_numeric, key_pool, -1);
 				pending_tile_features.clear();
+				if (outlayer->features.size() >= feature_limit) {
+					break;
+				}
 			}
 
 			if (preserve_input_order) {
@@ -1927,10 +1974,14 @@ std::string overzoom(std::vector<source_tile> const &tiles, int nz, int nx, int
 					std::vector<source_tile> sts;
 					sts.push_back(st);
 
+					// feature_limit arg is 1, because we just care whether there are any overzoomed features
+					// left after clipping to the child tile, not about their actual content
 					std::string child = overzoom(sts,
 								     nz + 1, nx * 2 + x, ny * 2 + y,
-								     detail, buffer, keep, exclude, exclude_prefix, false, NULL,
-								     demultiply, filter, preserve_input_order, attribute_accum, unidecode_data, simplification, tiny_polygon_size, bins, bin_by_id_list, accumulate_numeric);
+								     detail_or_unspecified, buffer, keep, exclude, exclude_prefix, false, NULL,
+								     demultiply, filter, preserve_input_order, attribute_accum, unidecode_data,
+								     simplification, tiny_polygon_size, bins, bin_by_id_list, accumulate_numeric,
+								     1);
 					if (child.size() > 0) {
 						next_overzoomed_tiles->emplace_back(nx * 2 + x, ny * 2 + y);
 					}
@@ -1940,8 +1991,9 @@ std::string overzoom(std::vector<source_tile> const &tiles, int nz, int nx, int
 	}
 
 	if (bins.size() > 0) {
-		outtile = assign_to_bins(outtile, bins, bin_by_id_list, nz, nx, ny, detail, attribute_accum, accumulate_numeric,
-					 keep, exclude, exclude_prefix);
+		outtile = assign_to_bins(outtile, bins, bin_by_id_list, nz, nx, ny,
+					 attribute_accum, accumulate_numeric,
+					 keep, exclude, exclude_prefix, buffer);
 	}
 
 	for (ssize_t i = outtile.layers.size() - 1; i >= 0; i--) {
